unit BS_LibraryManager;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  WinProcs,WinTypes,
  uBS_LibraryTypes
  ;

type
  ELibraryLoadError = class(Exception);
  TBS_LibraryManager = class;

  TBS_BeforeLibraryLoadEvent = procedure(Sender : TObject; aLibraryFilename : string; var aCanLoad: boolean) of object;
  TBS_AfterLibraryLoadEvent = procedure(Sender : TObject; aLibraryFilename : string; aLoadOk: boolean) of object;
  TBS_AfterLibraryUnLoadEvent = procedure(Sender : TObject; aLibraryFilename : string) of object;

  //Loaded Library collection item
  TLoadedLibraryItem = class(TCollectionItem)
  private
    FFileName             : String;
    FLibHandle            : THandle;
    FPckHandle            : HMODULE;
    FLibraryType          : TBS_LibraryType;
    FCollection           : TCollection;
    //   
    FLibraryInit          : TLibraryInit;
    FLibraryFinalize      : TLibraryFinalize;

    FGetLibraryInfo       : TGetLibraryInfo;      //   
    FLibraryRun           : TLibraryRun;          // 
    FLibraryRunSuspended  : TLibraryRunSuspended;
    FLibrarySuspend       : TLibrarySuspend;
    FLibraryResume        : TLibraryResume;

    FLibraryTerminate     : TLibraryTerminate;    // 
    FSetLibraryData       : TSetLibraryData;      //  
    FRegisterCallBack     : TRegisterCallBack;    // Register call back functions
  public
    constructor Create(Collection: TCollection); overload; override;
    destructor Destroy;override;

    procedure LibraryInit(aHostApplication:TApplication;aHostScreen:TScreen); // Library initialization
    procedure LibraryFinalize;                    // Library deinitialization

    function GetLibraryInfo:TLibraryInfo;         // Get library description
    function LibraryRun(const aClassName : PChar;
                        const crMode : Longint):Longint; // run
    //Same as LibraryRun, but Run suspended. Call LibraryResume to start exeqution
    function LibraryRunSuspended(const aClassName: PChar;
                                 const crMode: Integer): Longint;

    //Call LibrarySuspend to temporarily halt execution of the Library module.
    //To resume execution after a call to LibrarySuspend, call LibrarySuspendResume.
    procedure LibrarySuspend;
    //Call LibraryResume to cause a suspended library module to start running again.
    procedure LibraryResume;

    procedure LibraryTerminate;                        // stop library execution
    procedure SetLibraryData(aKind:TDataKind;
                             const aBuffer;
                             aSize:Integer);      //  
    procedure RegisterCallBack(aIndex:integer; aDddr: Pointer);    // Register call back functions
    property LibraryType : TBS_LibraryType read FLibraryType write FLibraryType;
  end;

  TLoadedLibries = class(TCollection)
  private
    FBS_LibraryManager       : TBS_LibraryManager;
    function GetItem(Index: Integer): TLoadedLibraryItem;
    procedure SetItem(Index: Integer; Value: TLoadedLibraryItem);
  public
    constructor Create(BS_LibraryManager: TBS_LibraryManager);
    destructor  Destroy;override;
    function    Add : TLoadedLibraryItem;
    property    Items[Index:integer]: TLoadedLibraryItem  read GetItem write SetItem;
  end;

  TBS_LibraryManager = class(TComponent)
  private
    FLoadedLibries        : TLoadedLibries;
    FOnBeforeLibraryLoad  : TBS_BeforeLibraryLoadEvent;
    FOnAfterLibraryLoad   : TBS_AfterLibraryLoadEvent;
    FOnAfterLibraryUnLoad : TBS_AfterLibraryUnLoadEvent;
  protected
    procedure BeforeLibraryLoad(aLibraryFilename : string; var aCanLoad : boolean);virtual;
    procedure AfterLibraryLoad(aLibraryFilename : string; aLoadOk : boolean);virtual;
    procedure AfterLibraryUnLoad(aLibraryFilename : string);virtual;
  public
    constructor Create(AOwner:TComponent); override;
    destructor Destroy; override;
    function  AddLibrary(const aFileName: String; const aLibraryType: TBS_LibraryType):boolean;
    procedure RemoveLibrary(aFileName:String); overload;
    procedure RemoveLibrary(aIndex:integer);   overload;
    function  GetLibraryIndex(aFileName:String):integer;
    property  LoadedLibries : TLoadedLibries read fLoadedLibries write fLoadedLibries;
  published
    property OnBeforeLibraryLoad : TBS_BeforeLibraryLoadEvent read FOnBeforeLibraryLoad write FOnBeforeLibraryLoad;
    property OnAfterLibraryLoad : TBS_AfterLibraryLoadEvent read FOnAfterLibraryLoad write FOnAfterLibraryLoad;
    property OnAfterLibraryUnLoad : TBS_AfterLibraryUnLoadEvent read FOnAfterLibraryUnLoad write FOnAfterLibraryUnLoad;
  end;


procedure LoadPackageInfoProc(const Name: string; NameType: TNameType; Flags: Byte; Param: Pointer);
procedure UnLoadPackageInfoProc(const Name: string; NameType: TNameType; Flags: Byte; Param: Pointer);


implementation
///////////////////////////////////////////////////////////////////////////////

//**
// TBS_LibraryManager
//
constructor TBS_LibraryManager.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  LoadedLibries := TLoadedLibries.Create(Self);
end;

destructor TBS_LibraryManager.Destroy;
var
 i:integer;
begin
  for i:=LoadedLibries.Count-1 downto 0 do
    RemoveLibrary(i);
  inherited Destroy;
end;

procedure TBS_LibraryManager.RemoveLibrary(aFileName: String);
var
  i:integer;
begin
  i:=GetLibraryIndex(aFileName);
  if i >= 0 then
    RemoveLibrary(i);
end;

procedure TBS_LibraryManager.RemoveLibrary(aIndex: integer);
begin
  Self.LoadedLibries.Delete(aIndex);
end;

function TBS_LibraryManager.GetLibraryIndex(aFileName: String): integer;
var
  i:integer;
begin
  Result:=-1; //Library not found
  if Assigned(Self) then
  for i:=Self.LoadedLibries.Count-1 downto 0 do
    if Self.LoadedLibries.Items[i].FFileName=aFileName then
    begin
      Result:=i;
      break;
    end;
end;

function TBS_LibraryManager.AddLibrary(const aFileName: String; const aLibraryType: TBS_LibraryType): boolean;
var
  libHandle   :THandle;
  pckHandle   :HMODULE;
  itemIndex   :integer;
  canLoadFlag :boolean;
  flags       :integer;
begin
  Result:=False;
  if GetLibraryIndex(aFileName) < 0 then
    case aLibraryType of
      ltDll:
       begin
         Self.BeforeLibraryLoad(aFileName,canLoadFlag);
         if canLoadFlag then
         try
          libHandle:=LoadLibrary(PChar(aFileName));
          if libHandle > 0 then
            begin
              itemIndex:=TCollectionItem(Self.LoadedLibries.Add).Index;
              Self.LoadedLibries.Items[itemIndex].FFileName:=aFileName;
              Self.LoadedLibries.Items[itemIndex].FLibHandle:=libHandle;
              Self.LoadedLibries.Items[itemIndex].FLibraryType:=ltDll;
              {@Self.LoadedLibries.Items[itemIndex].FLibraryInit:=GetProcAddress(libHandle,'LibraryInit');
              @Self.LoadedLibries.Items[itemIndex].FLibraryFinalize:=GetProcAddress(libHandle,'LibraryFinalize');
              @Self.LoadedLibries.Items[itemIndex].FLibraryRun:=GetProcAddress(libHandle,'LibraryRun');
              @Self.LoadedLibries.Items[itemIndex].FLibraryRunSuspended:=GetProcAddress(libHandle,'LibraryRunSuspended');
              @Self.LoadedLibries.Items[itemIndex].FLibrarySuspend:=GetProcAddress(libHandle,'LibrarySuspend');
              @Self.LoadedLibries.Items[itemIndex].FLibraryResume:=GetProcAddress(libHandle,'LibraryResume');
              @Self.LoadedLibries.Items[itemIndex].FLibraryTerminate:=GetProcAddress(libHandle,'LibraryTerminate');
              @Self.LoadedLibries.Items[itemIndex].FSetLibraryData:=GetProcAddress(libHandle,'SetLibraryData');
              @Self.LoadedLibries.Items[itemIndex].FRegisterCallBack:=GetProcAddress(libHandle,'RegisterCallBack');
              Self.LoadedLibries.Items[itemIndex].LibraryInit(Application,Screen);}


              Result:=True;
            end
           else
            ELibraryLoadError.Create('Unable to load DLL '+aFileName);
         finally
           Self.AfterLibraryLoad(aFileName,Result);
         end;
       end;
      ltBpl:
      begin
        Self.BeforeLibraryLoad(aFileName,canLoadFlag);
        if canLoadFlag then
        try
          pckHandle:=LoadPackage(aFileName);
          if pckHandle > 0 then
            begin
              itemIndex:=TCollectionItem(Self.LoadedLibries.Add).Index;
              Self.LoadedLibries.Items[itemIndex].FFileName:=aFileName;
              Self.LoadedLibries.Items[itemIndex].FPckHandle:=pckHandle;
              Self.LoadedLibries.Items[itemIndex].FLibraryType:=ltBpl;
              {@Self.LoadedLibries.Items[itemIndex].FLibraryInit:=GetProcAddress(libHandle,'LibraryInit');
              @Self.LoadedLibries.Items[itemIndex].FLibraryFinalize:=GetProcAddress(libHandle,'LibraryFinalize');
              @Self.LoadedLibries.Items[itemIndex].FLibraryRun:=GetProcAddress(libHandle,'LibraryRun');
              @Self.LoadedLibries.Items[itemIndex].FLibraryRunSuspended:=GetProcAddress(libHandle,'LibraryRunSuspended');
              @Self.LoadedLibries.Items[itemIndex].FLibrarySuspend:=GetProcAddress(libHandle,'LibrarySuspend');
              @Self.LoadedLibries.Items[itemIndex].FLibraryResume:=GetProcAddress(libHandle,'LibraryResume');
              @Self.LoadedLibries.Items[itemIndex].FLibraryTerminate:=GetProcAddress(libHandle,'LibraryTerminate');
              @Self.LoadedLibries.Items[itemIndex].FSetLibraryData:=GetProcAddress(libHandle,'SetLibraryData');
              @Self.LoadedLibries.Items[itemIndex].FRegisterCallBack:=GetProcAddress(libHandle,'RegisterCallBack');
              Self.LoadedLibries.Items[itemIndex].LibraryInit(Application,Screen); }
              GetPackageInfo(pckHandle, Pointer(pckHandle),
                             Flags, LoadPackageInfoProc);

              Result:=True;
            end
          else
            ELibraryLoadError.Create('Unable to load Bpl '+aFileName);
        finally
          Self.AfterLibraryLoad(aFileName,Result);
        end;
      end;
      else
        Result:=False;
    end //case
  else
    Result:=True; //The library is already loaded
end;

procedure TBS_LibraryManager.BeforeLibraryLoad(aLibraryFilename: string; var aCanLoad: boolean);
begin
  aCanLoad := True;
  if assigned(FOnBeforeLibraryLoad) then
      FOnBeforeLibraryLoad(Self, aLibraryFilename, aCanLoad);
end;

procedure TBS_LibraryManager.AfterLibraryLoad(aLibraryFilename: string; aLoadOk : boolean);
begin
  if assigned(FOnAfterLibraryLoad) then
    FOnAfterLibraryLoad(Self, aLibraryFilename, aLoadOk);
end;

procedure TBS_LibraryManager.AfterLibraryUnLoad(aLibraryFilename: string);
begin
  if assigned(FOnAfterLibraryUnLoad) then
    FOnAfterLibraryUnLoad(Self, aLibraryFilename);
end;


//**
// TLoadedLibries
//
constructor TLoadedLibries.Create(BS_LibraryManager: TBS_LibraryManager);
begin
  inherited Create(TLoadedLibraryItem);
  Self.FBS_LibraryManager:=BS_LibraryManager;  //Collection owner
end;

destructor TLoadedLibries.Destroy;
begin
  Self.Clear;
  inherited Destroy;
end;

function TLoadedLibries.Add: TLoadedLibraryItem;
begin
  Result:=TLoadedLibraryItem(inherited Add);
end;

function TLoadedLibries.GetItem(Index: Integer): TLoadedLibraryItem;
begin
  Result:=TLoadedLibraryItem(inherited GetItem(Index));
end;

procedure TLoadedLibries.SetItem(Index: Integer;
  Value: TLoadedLibraryItem);
begin
  inherited SetItem(Index, Value);
end;


//**
// TLoadedLibraryItem
//
constructor TLoadedLibraryItem.Create(Collection: TCollection);
begin
  inherited Create(Collection);
  Self.FCollection:=Collection;
  Self.FLibHandle:=0;
  Self.FPckHandle:=0;
  Self.FFileName:='';
end;

destructor TLoadedLibraryItem.Destroy;
var
  flags : Integer;
begin
  LibraryTerminate; //Terminate if assigned
  LibraryFinalize;
  case Self.FLibraryType of
    ltDll: if Self.FLibHandle<>0 then
            FreeLibrary(Self.FLibHandle); //Unload dll
    ltBpl:if Self.FPckHandle<>0 then
            begin
              try //Unregister package
                flags := 31;
                GetPackageInfo(Self.FPckHandle, Pointer(Self.FPckHandle),
                               flags, UnLoadPackageInfoProc);
              except
                //on e: Exception do
                //  Application.MessageBox(PChar(e.Message), PChar(sError), MB_OK + MB_ICONWARNING);
              end;
              UnloadPackage(Self.FPckHandle); //Unload dllpckHandle
              TLoadedLibries(Self.FCollection).FBS_LibraryManager.FOnAfterLibraryUnLoad(Self,Self.FFileName);
            end;
  end;
  inherited Destroy;
end;

// Library initialization
procedure TLoadedLibraryItem.LibraryInit(aHostApplication:TApplication;aHostScreen:TScreen);
begin
  if @FLibraryInit <> nil then
    FLibraryInit(aHostApplication,aHostScreen)
  else
    if Self.FLibraryType = ltDll then
      raise Exception.CreateFmt('LibraryInit method in %s is not assigned',[Self.FFileName]);
end;

procedure TLoadedLibraryItem.LibraryFinalize;
begin
  if @FLibraryFinalize <> nil then
    FLibraryFinalize
  else
    if Self.FLibraryType = ltDll then
      raise Exception.CreateFmt('Finalize method in %s is not assigned',[Self.FFileName]);
end;

function TLoadedLibraryItem.GetLibraryInfo: TLibraryInfo;
begin
  if @FGetLibraryInfo<>nil then
    Result:=FGetLibraryInfo;
end;

function TLoadedLibraryItem.LibraryRun(const aClassName : PChar;
                                       const crMode : Longint):Longint;
begin
  if @FLibraryRun<>nil then
    Result:=FLibraryRun(aClassName, crMode)
  else
    Result:=-1;
end;

function TLoadedLibraryItem.LibraryRunSuspended(
  const aClassName: PChar;
  const crMode: Integer): Longint;
begin
  if @FLibraryRunSuspended<>nil then
    Result:=FLibraryRunSuspended(aClassName, crMode)
  else
    Result:=-1;
end;

procedure TLoadedLibraryItem.SetLibraryData(aKind: TDataKind; const aBuffer;
  aSize: Integer);
begin
  if @FSetLibraryData<>nil then
    FSetLibraryData(aKind,aBuffer,aSize);
end;

procedure TLoadedLibraryItem.LibraryTerminate;
begin
  if @FLibraryTerminate <> nil then
    FLibraryTerminate;
end;

procedure TLoadedLibraryItem.RegisterCallBack(aIndex: integer;
  aDddr: Pointer);
begin
  if @FRegisterCallBack <> nil then
    FRegisterCallBack(aIndex, aDddr);
end;

procedure TLoadedLibraryItem.LibraryResume;
begin
  if @FLibraryResume <> nil then
    FLibraryResume;
end;

procedure TLoadedLibraryItem.LibrarySuspend;
begin
  if @FLibrarySuspend <> nil then
    FLibrarySuspend;
end;


procedure LoadPackageInfoProc(const Name: string; NameType: TNameType; Flags: Byte; Param: Pointer);
type
   TRegisterProc = procedure;
var
   RegisterProc: TRegisterProc;
   localName: String;
begin
   if NameType = ntContainsUnit then
   begin
      localName := LowerCase(Name);
      if Length(localName) > 0 then
      localName[1] := UpCase(localName[1]);
      localName:='@' + localName + '@Register$qqrv';
      @RegisterProc := GetProcAddress(HModule(Param), PChar(localName));

      if @RegisterProc <> nil then
      begin
         RegisterProc;
      end
      else
      begin
         // contains unit
      end;
   end
   else
   begin
      // Requires package
   end;
end;

procedure UnLoadPackageInfoProc(const Name: string; NameType: TNameType; Flags: Byte; Param: Pointer);
type
   TUnRegisterProc = procedure;
var
   UnRegisterProc: TUnRegisterProc;
   localName: String;
begin
   if NameType = ntContainsUnit then
   begin
      localName := LowerCase(Name);
      if Length(localName) > 0 then
      localName[1] := UpCase(localName[1]);
      localName:='@' + localName + '@UnRegister$qqrv';
      @UnRegisterProc := GetProcAddress(HModule(Param), PChar(localName));

      if @UnRegisterProc <> nil then
      begin
         UnRegisterProc;
      end
      else
      begin
         // contains unit
      end;
   end
   else
   begin
      // Requires package
   end;
end;		{ PackageIsLoadingProc }



end.
